"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SimpleClient = exports.Ohlcv = exports.Ticker = exports.FetchMarket = exports.FetchMarketSymbol = exports.Balance = exports.MarginAccountBalance = exports.MarketBalance = void 0;
const web3_js_1 = require("@solana/web3.js");
const os_1 = __importDefault(require("os"));
const fs_1 = __importDefault(require("fs"));
const serum_1 = require("@project-serum/serum");
const utils_1 = require("./utils");
const bn_js_1 = __importDefault(require("bn.js"));
const node_fetch_1 = __importDefault(require("node-fetch"));
const mango_client_1 = require("@blockworks-foundation/mango-client");
class MarketBalance {
    constructor(baseTokenSymbol, orders, unsettled, quoteTokenSymbol, quoteOrders, quoteUnsettled) {
        this.baseTokenSymbol = baseTokenSymbol;
        this.orders = orders;
        this.unsettled = unsettled;
        this.quoteTokenSymbol = quoteTokenSymbol;
        this.quoteOrders = quoteOrders;
        this.quoteUnsettled = quoteUnsettled;
    }
}
exports.MarketBalance = MarketBalance;
class MarginAccountBalance {
    constructor(tokenSymbol, deposited, borrowed) {
        this.tokenSymbol = tokenSymbol;
        this.deposited = deposited;
        this.borrowed = borrowed;
    }
}
exports.MarginAccountBalance = MarginAccountBalance;
class Balance {
    constructor(marginAccountPublicKey, marginAccountBalances, marketBalances) {
        this.marginAccountPublicKey = marginAccountPublicKey;
        this.marginAccountBalances = marginAccountBalances;
        this.marketBalances = marketBalances;
    }
}
exports.Balance = Balance;
class FetchMarketSymbol {
    constructor(symbol) {
        this.symbol = symbol;
    }
}
exports.FetchMarketSymbol = FetchMarketSymbol;
class FetchMarket {
    constructor(symbols) {
        this.symbols = symbols;
    }
}
exports.FetchMarket = FetchMarket;
class Ticker {
    constructor(symbol, price, timeMs) {
        this.symbol = symbol;
        this.price = price;
        this.timeMs = timeMs;
    }
}
exports.Ticker = Ticker;
class Ohlcv {
    constructor(timeS, open, high, low, close, volume) {
        this.timeS = timeS;
        this.open = open;
        this.high = high;
        this.low = low;
        this.close = close;
        this.volume = volume;
    }
}
exports.Ohlcv = Ohlcv;
class EmptyOrderBookError extends Error {
    constructor(message) {
        super(message);
        this.name = 'EmptyOrderBookError';
    }
}
/**
 * a simpler more cex-style client with sensible (hopefully ;)) defaults
 */
class SimpleClient {
    constructor(client, connection, programId, dexProgramId, mangoGroup, markets, mangoGroupTokenMappings, spotMarketMappings, payer, marginAccountPk) {
        this.client = client;
        this.connection = connection;
        this.programId = programId;
        this.dexProgramId = dexProgramId;
        this.mangoGroup = mangoGroup;
        this.markets = markets;
        this.mangoGroupTokenMappings = mangoGroupTokenMappings;
        this.spotMarketMappings = spotMarketMappings;
        this.payer = payer;
        this.marginAccountPk = marginAccountPk;
    }
    static async create(marginAccountPk) {
        const cluster = process.env.CLUSTER || 'mainnet-beta';
        const mangoGroupName = 'BTC_ETH_SOL_SRM_USDC';
        const clusterIds = mango_client_1.IDS[cluster];
        const programId = new web3_js_1.PublicKey(mango_client_1.IDS[cluster].mango_program_id);
        const dexProgramId = new web3_js_1.PublicKey(clusterIds.dex_program_id);
        const mangoGroupIds = clusterIds.mango_groups[mangoGroupName];
        // connection
        const connection = new web3_js_1.Connection(mango_client_1.IDS.cluster_urls[cluster], 'processed');
        // client
        const client = new mango_client_1.MangoClient();
        // payer
        const keyPairPath = process.env.KEYPAIR || os_1.default.homedir() + '/.config/solana/id.json';
        const payer = new web3_js_1.Account(JSON.parse(fs_1.default.readFileSync(keyPairPath, 'utf-8')));
        // mangoGroup
        const mangoGroupPk = new web3_js_1.PublicKey(clusterIds.mango_groups[mangoGroupName].mango_group_pk);
        const mangoGroup = await client.getMangoGroup(connection, mangoGroupPk);
        // markets
        const markets = await Promise.all(mangoGroup.spotMarkets.map((pk) => serum_1.Market.load(connection, pk, { skipPreflight: true, commitment: 'singleGossip' }, dexProgramId)));
        // token mappings
        const mangoGroupTokenMappings = new Map();
        const mangoGroupSymbols = Object.entries(mangoGroupIds.symbols);
        for (const [tokenName, tokenMint] of mangoGroupSymbols) {
            mangoGroupTokenMappings[tokenName] = new web3_js_1.PublicKey(tokenMint);
        }
        // market mappings
        const mangoGroupSportMarketMappings = new Map();
        const mangoGroupSpotMarketSymbols = Object.entries(mangoGroupIds.spot_market_symbols);
        for (const [spotMarketSymbol, address] of mangoGroupSpotMarketSymbols) {
            mangoGroupSportMarketMappings[spotMarketSymbol] = new web3_js_1.PublicKey(address);
        }
        return new SimpleClient(client, connection, programId, dexProgramId, mangoGroup, markets, mangoGroupTokenMappings, mangoGroupSportMarketMappings, payer, marginAccountPk);
    }
    /// private
    getMarketForSymbol(marketSymbol) {
        if (Object.keys(this.spotMarketMappings).indexOf(marketSymbol) === -1) {
            throw new Error(`unknown spot market ${marketSymbol}`);
        }
        const marketAddress = this.spotMarketMappings[marketSymbol];
        const market = this.markets.find((market) => market.publicKey.equals(marketAddress));
        if (market === undefined) {
            throw new Error(`market not found for ${market}`);
        }
        return market;
    }
    async getMarginAccountForOwner() {
        return await this.client.getMarginAccount(this.connection, new web3_js_1.PublicKey(this.marginAccountPk), this.dexProgramId);
    }
    async getOpenOrdersAccountForSymbol(marketSymbol) {
        const market = this.getMarketForSymbol(marketSymbol);
        const marketIndex = this.mangoGroup.getMarketIndex(market);
        const marginAccount = await this.getMarginAccountForOwner();
        return marginAccount.openOrdersAccounts[marketIndex];
    }
    async cancelOrder(marginAccount, market, order) {
        return this.client.cancelOrder(this.connection, this.programId, this.mangoGroup, marginAccount, this.payer, market, order);
    }
    async cancelOrdersForMarginAccount(marginAccount, symbol, clientId) {
        let orders;
        let market;
        if (symbol === undefined) {
            for (const spotMarketSymbol of Object.keys(this.spotMarketMappings)) {
                market = this.getMarketForSymbol(spotMarketSymbol);
                orders = await this.getOpenOrders(spotMarketSymbol);
                await orders.map((order) => this.cancelOrder(marginAccount, market, order));
            }
            return;
        }
        market = this.getMarketForSymbol(symbol);
        orders = await this.getOpenOrders(symbol);
        // note: clientId could not even belong to his margin account
        // in that case ordersToCancel would be empty
        const ordersToCancel = clientId !== undefined
            ? orders.filter((o) => o.clientId.toString() === clientId)
            : orders;
        await Promise.all(ordersToCancel.map((order) => this.cancelOrder(marginAccount, market, order)));
    }
    /// public
    async placeOrder(symbol, type, side, quantity, price, orderType) {
        if (!symbol.trim()) {
            throw new Error(`invalid symbol ${symbol}`);
        }
        if (!Number.isFinite(quantity) || quantity <= 0) {
            throw new Error(`invalid quantity ${quantity}`);
        }
        if ((type === 'limit' && !Number.isFinite(price)) || price <= 0) {
            throw new Error(`invalid price ${price}`);
        }
        if (type === 'market') {
            const orderBook = await this.getOrderBook(symbol);
            let acc = 0;
            let selectedOrder;
            if (orderBook.length === 0) {
                throw new EmptyOrderBookError('Empty order book encountered when placing a market order!');
            }
            for (const order of orderBook) {
                acc += order.size;
                if (acc >= quantity) {
                    selectedOrder = order;
                    break;
                }
            }
            if (side === 'buy') {
                price = selectedOrder.price * 1.05;
            }
            else {
                price = selectedOrder.price * 0.95;
            }
        }
        const market = this.getMarketForSymbol(symbol);
        const marginAccount = await this.getMarginAccountForOwner();
        const clientId = new bn_js_1.default(Date.now());
        orderType = orderType === undefined ? 'limit' : orderType;
        await this.client.placeOrder(this.connection, this.programId, this.mangoGroup, marginAccount, market, this.payer, side, price, quantity, orderType, clientId);
        return clientId.toString();
    }
    async getOpenOrders(symbol, clientId) {
        const openOrderAccount = await this.getOpenOrdersAccountForSymbol(symbol);
        if (openOrderAccount === undefined) {
            return [];
        }
        let orders = await this.getOrderBook(symbol);
        orders = orders.filter((o) => openOrderAccount.address.equals(o.openOrdersAddress));
        if (clientId) {
            return orders.filter((o) => o.clientId && o.clientId.toString() === clientId);
        }
        return orders;
    }
    async cancelOrders(symbol, clientId) {
        const marginAccount = await this.getMarginAccountForOwner();
        await this.cancelOrdersForMarginAccount(marginAccount, symbol, clientId);
    }
    async getTradeHistory(symbol) {
        if (!symbol.trim()) {
            throw new Error(`invalid symbol ${symbol}`);
        }
        const openOrdersAccount = await this.getOpenOrdersAccountForSymbol(symbol);
        if (openOrdersAccount === undefined) {
            return [];
        }
        // e.g. https://stark-fjord-45757.herokuapp.com/trades/open_orders/G5rZ4Qfv5SxpJegVng5FuZftDrJkzLkxQUNjEXuoczX5
        //     {
        //         "id": 2267328,
        //         "loadTimestamp": "2021-04-28T03:36:20.573Z",
        //         "address": "C1EuT9VokAKLiW7i2ASnZUvxDoKuKkCpDDeNxAptuNe4",
        //         "programId": "9xQeWvG816bUx9EPjHmaT23yvVM2ZWbrrpZb9PusVFin",
        //         "baseCurrency": "BTC",
        //         "quoteCurrency": "USDT",
        //         "fill": true,
        //         "out": false,
        //         "bid": false,
        //         "maker": false,
        //         "openOrderSlot": "6",
        //         "feeTier": "0",
        //         "nativeQuantityReleased": "93207112",
        //         "nativeQuantityPaid": "1700",
        //         "nativeFeeOrRebate": "205508",
        //         "orderId": "9555524110645989995606320",
        //         "openOrders": "G5rZ4Qfv5SxpJegVng5FuZftDrJkzLkxQUNjEXuoczX5",
        //         "clientOrderId": "0",
        //         "uuid": "0040cdbdb0667fd5f75c2538e4097c5090e7b15d8cf9a5e7db7a54c3c212d27a",
        //         "source": "1",
        //         "baseTokenDecimals": 6,
        //         "quoteTokenDecimals": 6,
        //         "side": "sell",
        //         "price": 54948.6,
        //         "feeCost": 0.205508,
        //         "size": 0.0017
        //     }
        const response = await node_fetch_1.default(`https://stark-fjord-45757.herokuapp.com/trades/open_orders/${openOrdersAccount.address.toBase58()}`);
        const parsedResponse = await response.json();
        const trades = (parsedResponse === null || parsedResponse === void 0 ? void 0 : parsedResponse.data)
            ? parsedResponse.data
            : [];
        return trades
            .filter((trade) => openOrdersAccount.address.equals(new web3_js_1.PublicKey(trade.openOrders)))
            .map((trade) => ({ ...trade, marketName: symbol }));
    }
    /**
     * returns balances, simple and mango specific details
     */
    async getBalance() {
        const marginAccount = await this.getMarginAccountForOwner();
        const marginAccountBalances = Object.keys(this.mangoGroupTokenMappings).map((tokenSymbol) => {
            const tokenIndex = this.mangoGroup.getTokenIndex(this.mangoGroupTokenMappings[tokenSymbol]);
            const decimals = this.mangoGroup.mintDecimals[tokenIndex];
            const uiDeposit = marginAccount.getUiDeposit(this.mangoGroup, tokenIndex);
            const uiDepositDisplay = utils_1.ceilToDecimal(uiDeposit, decimals);
            const uiBorrow = marginAccount.getUiBorrow(this.mangoGroup, tokenIndex);
            const uiBorrowDisplay = utils_1.ceilToDecimal(uiBorrow, decimals);
            return new MarginAccountBalance(tokenSymbol, uiDepositDisplay, uiBorrowDisplay);
        });
        const marketBalances = this.mangoGroup.spotMarkets.map((marketPk) => {
            const market = this.markets.find((market) => market.publicKey.equals(marketPk));
            if (market === undefined) {
                throw new Error(`market for ${marketPk.toBase58()} not found`);
            }
            const token = Object.entries(this.mangoGroupTokenMappings).find((entry) => {
                return entry[1].equals(market.baseMintAddress);
            });
            const tokenIndex = this.mangoGroup.getTokenIndex(market.baseMintAddress);
            const openOrders = marginAccount.openOrdersAccounts[tokenIndex];
            const nativeBaseFree = (openOrders === null || openOrders === void 0 ? void 0 : openOrders.baseTokenFree) || new bn_js_1.default(0);
            const nativeBaseLocked = openOrders
                ? openOrders.baseTokenTotal.sub(nativeBaseFree)
                : new bn_js_1.default(0);
            const nativeBaseUnsettled = (openOrders === null || openOrders === void 0 ? void 0 : openOrders.baseTokenFree) || new bn_js_1.default(0);
            const orders = utils_1.nativeToUi(nativeBaseLocked.toNumber(), this.mangoGroup.mintDecimals[tokenIndex]);
            const unsettled = utils_1.nativeToUi(nativeBaseUnsettled.toNumber(), this.mangoGroup.mintDecimals[tokenIndex]);
            const quoteToken = Object.entries(this.mangoGroupTokenMappings).find((entry) => {
                return entry[1].equals(market.quoteMintAddress);
            });
            const quoteCurrencyIndex = this.mangoGroup.getTokenIndex(market.quoteMintAddress);
            const nativeQuoteFree = (openOrders === null || openOrders === void 0 ? void 0 : openOrders.quoteTokenFree) || new bn_js_1.default(0);
            const nativeQuoteLocked = openOrders
                ? openOrders.quoteTokenTotal.sub(nativeQuoteFree)
                : new bn_js_1.default(0);
            const nativeQuoteUnsettled = (openOrders === null || openOrders === void 0 ? void 0 : openOrders.quoteTokenFree) || new bn_js_1.default(0);
            const ordersQuote = utils_1.nativeToUi(nativeQuoteLocked.toNumber(), this.mangoGroup.mintDecimals[quoteCurrencyIndex]);
            const unsettledQuote = utils_1.nativeToUi(nativeQuoteUnsettled.toNumber(), this.mangoGroup.mintDecimals[quoteCurrencyIndex]);
            return new MarketBalance(token[0], orders, unsettled, quoteToken[0], ordersQuote, unsettledQuote);
        });
        return new Balance(marginAccount.publicKey.toBase58(), marginAccountBalances, marketBalances);
    }
    async getPnl() {
        // grab trade history
        let tradeHistory = [];
        for (const [spotMarketSymbol, unused] of Object.entries(this.spotMarketMappings)) {
            const tradeHistoryForSymbol = await this.getTradeHistory(spotMarketSymbol);
            tradeHistory = tradeHistory.concat(tradeHistoryForSymbol);
        }
        const profitAndLoss = {};
        // compute profit and loss for all markets
        const groupedTrades = utils_1.groupBy(tradeHistory, (trade) => trade.marketName);
        groupedTrades.forEach((val, key) => {
            profitAndLoss[key] = val.reduce((acc, current) => (current.side === 'sell' ? current.size * -1 : current.size) + acc, 0);
        });
        // compute profit and loss for usdc
        const totalNativeUsdc = tradeHistory.reduce((acc, current) => {
            const usdcAmount = current.side === 'sell'
                ? current.nativeQuantityReleased
                : current.nativeQuantityPaid * -1;
            return usdcAmount + acc;
        }, 0);
        profitAndLoss.USDC = utils_1.nativeToUi(totalNativeUsdc, this.mangoGroup.mintDecimals[2]);
        // compute final pnl
        let total = 0;
        const prices = await this.mangoGroup.getPrices(this.connection);
        const assetIndex = {
            'BTC/USDC': 0,
            'BTC/WUSDC': 0,
            'ETH/USDC': 1,
            'ETH/WUSDC': 1,
            'SOL/USDC': 1,
            'SOL/WUSDC': 1,
            'SRM/USDC': 1,
            'SRM/WUSDC': 1,
            USDC: 2,
            WUSDC: 2,
        };
        for (const assetName of Object.keys(profitAndLoss)) {
            total = total + profitAndLoss[assetName] * prices[assetIndex[assetName]];
        }
        return total.toFixed(2);
    }
    /**
     * returns available markets
     */
    async getMarkets() {
        const fetchMarketSymbols = Object.keys(this.spotMarketMappings).map((spotMarketSymbol) => new FetchMarketSymbol(spotMarketSymbol));
        return new FetchMarket(fetchMarketSymbols);
    }
    /**
     * returns tickers i.e. symbol, closing price, time of closing price
     */
    async getTickers(symbol) {
        let ohlcvs;
        let latestOhlcv;
        const to = Date.now();
        // use a sufficiently large window to ensure that we get data back
        const toMinus20Mins = to - 20 * 60 * 1000;
        const oneMinute = '1';
        if (symbol === undefined) {
            const tickers = [];
            for (const zymbol of Object.keys(this.spotMarketMappings)) {
                ohlcvs = await this.getOhlcv(zymbol, oneMinute, toMinus20Mins, to);
                latestOhlcv = ohlcvs[ohlcvs.length - 1];
                tickers.push(new Ticker(zymbol, latestOhlcv.close, latestOhlcv.timeS * 1000));
            }
            return tickers;
        }
        ohlcvs = await this.getOhlcv(symbol, oneMinute, toMinus20Mins, to);
        latestOhlcv = ohlcvs[ohlcvs.length - 1];
        return [new Ticker(symbol, latestOhlcv.close, latestOhlcv.timeS * 1000)];
    }
    async getOrderBook(symbol) {
        var _a, _b;
        const market = this.getMarketForSymbol(symbol);
        const bidData = (_a = (await this.connection.getAccountInfo(market.bidsAddress))) === null || _a === void 0 ? void 0 : _a.data;
        const bidOrderBook = bidData
            ? serum_1.Orderbook.decode(market, Buffer.from(bidData))
            : [];
        const askData = (_b = (await this.connection.getAccountInfo(market.asksAddress))) === null || _b === void 0 ? void 0 : _b.data;
        const askOrderBook = askData
            ? serum_1.Orderbook.decode(market, Buffer.from(askData))
            : [];
        return [...bidOrderBook, ...askOrderBook];
    }
    /**
     * returns ohlcv in ascending order for time
     */
    async getOhlcv(spotMarketSymbol, resolution, fromEpochMs, toEpochMs) {
        const response = await node_fetch_1.default(`https://serum-history.herokuapp.com/tv/history` +
            `?symbol=${spotMarketSymbol}&resolution=${resolution}` +
            `&from=${fromEpochMs / 1000}&to=${toEpochMs / 1000}`);
        const { t, o, h, l, c, v } = await response.json();
        const ohlcvs = [];
        for (let i = 0; i < t.length; i++) {
            ohlcvs.push(new Ohlcv(t[i], o[i], h[i], l[i], c[i], v[i]));
        }
        return ohlcvs;
    }
}
exports.SimpleClient = SimpleClient;
//# sourceMappingURL=simpleclient.js.map